@@ -121,14 +121,17 @@ Load(SV *yaml_sv)
     dXSARGS;
     perl_yaml_loader_t loader;
     SV *node;
-    char *yaml_str;
+    const unsigned char *yaml_str;
     STRLEN yaml_len;
-    
-    /* If UTF8, make copy and downgrade */
-    if (SvPV_nolen(yaml_sv) && SvUTF8(yaml_sv)) {
+
+    yaml_str = (const unsigned char *)SvPV_const(yaml_sv, yaml_len);
+
+    if (DO_UTF8(yaml_sv)) {
         yaml_sv = sv_mortalcopy(yaml_sv);
+        if (!sv_utf8_downgrade(yaml_sv, TRUE))
+            croak("Wide character in YAML::XS::Load()");
+        yaml_str = (const unsigned char *)SvPV_const(yaml_sv, yaml_len);
     }
-    yaml_str = SvPVbyte(yaml_sv, yaml_len);
 
     sp = mark;
     if (0 && (items || ax)) {} /* XXX Quiet the -Wall warnings for now. */
@@ -137,7 +140,7 @@ Load(SV *yaml_sv)
     loader.document = 0;
     yaml_parser_set_input_string(
         &loader.parser,
-        (unsigned char *)yaml_str,
+        yaml_str,
         yaml_len
     );
 
@@ -151,16 +154,20 @@ Load(SV *yaml_sv)
          );
 
     loader.anchors = newHV();
-    sv_2mortal(loader.anchors);
+    sv_2mortal((SV *)loader.anchors);
 
     /* Keep calling load_node until end of stream */
     while (1) {
         loader.document++;
+        /* We are through with the previous event - delete it! */
+        yaml_event_delete(&loader.event);
         if (!yaml_parser_parse(&loader.parser, &loader.event))
             goto load_error;
         if (loader.event.type == YAML_STREAM_END_EVENT)
             break;
         node = load_node(&loader);
+        /* We are through with the previous event - delete it! */
+        yaml_event_delete(&loader.event);
         hv_clear(loader.anchors);
         if (! node) break;
         XPUSHs(sv_2mortal(node));
@@ -190,52 +197,81 @@ load_error:
 SV *
 load_node(perl_yaml_loader_t *loader)
 {
+    SV* return_sv = NULL;
+    /* This uses stack, but avoids (severe!) memory leaks */
+    yaml_event_t uplevel_event;
+
+    uplevel_event = loader->event;
+
     /* Get the next parser event */
     if (!yaml_parser_parse(&loader->parser, &loader->event))
         goto load_error;
 
+    /* These events don't need yaml_event_delete */
+    /* Some kind of error occurred */
+    if (loader->event.type == YAML_NO_EVENT)
+        goto load_error;
+
     /* Return NULL when we hit the end of a scope */
     if (loader->event.type == YAML_DOCUMENT_END_EVENT ||
         loader->event.type == YAML_MAPPING_END_EVENT ||
-        loader->event.type == YAML_SEQUENCE_END_EVENT) return NULL;
-
-    /* Handle loading a mapping */
-    if (loader->event.type == YAML_MAPPING_START_EVENT) {
-        SV *hash_ref;
-        char *tag = (char *)loader->event.data.mapping_start.tag;
-
-        /* Handle mapping tagged as a Perl hard reference */
-        if (tag && strEQ(tag, TAG_PERL_REF))
-            return load_scalar_ref(loader);
-        
-        /* Handle mapping tagged as a Perl typeglob */
-        if (tag && strEQ(tag, TAG_PERL_GLOB))
-            return load_glob(loader);
-
-        /* Load the mapping into a hash ref and return it */
-        return load_mapping(loader, NULL);
+        loader->event.type == YAML_SEQUENCE_END_EVENT) {
+            /* restore the uplevel event, so it can be properly deleted */
+            loader->event = uplevel_event;
+            return return_sv;
     }
 
-    /* Handle loading a sequence into an array */
-    if (loader->event.type == YAML_SEQUENCE_START_EVENT)
-        return load_sequence(loader);
+    /* The rest all need cleanup */
+    switch (loader->event.type) {
+        char *tag;
 
-    /* Handle loading a scalar */
-    if (loader->event.type == YAML_SCALAR_EVENT)
-        return load_scalar(loader);
+        /* Handle loading a mapping */
+        case YAML_MAPPING_START_EVENT:
+            tag = (char *)loader->event.data.mapping_start.tag;
 
-    /* Handle loading an alias node */
-    if (loader->event.type == YAML_ALIAS_EVENT)
-        return load_alias(loader);
+            /* Handle mapping tagged as a Perl hard reference */
+            if (tag && strEQ(tag, TAG_PERL_REF)) {
+                return_sv = load_scalar_ref(loader);
+                break;
+            }
 
-    /* Some kind of error occurred */
-    if (loader->event.type == YAML_NO_EVENT)
+            /* Handle mapping tagged as a Perl typeglob */
+            if (tag && strEQ(tag, TAG_PERL_GLOB)) {
+                return_sv = load_glob(loader);
+                break;
+            }
+
+            return_sv = load_mapping(loader, NULL);
+            break;
+
+        /* Handle loading a sequence into an array */
+        case YAML_SEQUENCE_START_EVENT:
+            return_sv = load_sequence(loader);
+            break;
+
+        /* Handle loading a scalar */
+        case YAML_SCALAR_EVENT:
+            return_sv = load_scalar(loader);
+            break;
+
+        /* Handle loading an alias node */
+        case YAML_ALIAS_EVENT:
+            return_sv = load_alias(loader);
+            break;
+
+        default:
+            croak(ERRMSG "Invalid event '%d' at top level", (int) loader->event.type);
+    }
+
+    yaml_event_delete(&loader->event);
+
+    /* restore the uplevel event, so it can be properly deleted */
+    loader->event = uplevel_event;
+
+    return return_sv;
+
+    load_error:
         croak(loader_error_msg(loader, NULL));
-
-    croak(ERRMSG "Invalid event '%d' at top level", (int) loader->event.type);
-
-load_error:
-    croak(loader_error_msg(loader, NULL));
 }
 
 /*
@@ -262,7 +298,7 @@ load_mapping(perl_yaml_loader_t *loader, char *tag)
         assert(SvPOK(key_node));
         value_node = load_node(loader);
         hv_store_ent(
-            hash, key_node, value_node, 0
+            hash, sv_2mortal(key_node), value_node, 0
         );
     } 
 
@@ -342,7 +378,7 @@ load_scalar(perl_yaml_loader_t *loader)
         class = tag + strlen(prefix);
         scalar = sv_setref_pvn(newSV(0), class, string, strlen(string));
         SvUTF8_on(scalar);
-	return scalar;
+    return scalar;
     }
 
     if (loader->event.data.scalar.style == YAML_PLAIN_SCALAR_STYLE) {
@@ -350,6 +386,8 @@ load_scalar(perl_yaml_loader_t *loader)
             return newSV(0);
         else if (strEQ(string, ""))
             return newSV(0);
+        else if (strEQ(string, "null"))
+            return newSV(0);
         else if (strEQ(string, "true"))
             return &PL_sv_yes;
         else if (strEQ(string, "false"))
@@ -363,7 +401,7 @@ load_scalar(perl_yaml_loader_t *loader)
         SvIV_please(scalar);
     }
 
-    SvUTF8_on(scalar);
+    (void)sv_utf8_decode(scalar);
     if (anchor)
         hv_store(loader->anchors, anchor, strlen(anchor), SvREFCNT_inc(scalar), 0);
     return scalar;
@@ -463,6 +501,12 @@ set_dumper_options(perl_yaml_dumper_t *dumper)
         ((gv = gv_fetchpv("YAML::XS::DumpCode", TRUE, SVt_PV)) &&
         SvTRUE(GvSV(gv)))
     );
+
+    dumper->quote_number_strings = (
+        ((gv = gv_fetchpv("YAML::XS::QuoteNumericStrings", TRUE, SVt_PV)) &&
+        SvTRUE(GvSV(gv)))
+    );
+    // dumper->emitter.open_ended = 1;
 }
 
 /*
@@ -574,10 +618,9 @@ dump_prewalk(perl_yaml_dumper_t *dumper, SV *node)
     }
     else if (ref_type == SVt_PVHV) {
         HV *hash = (HV *)SvRV(node);
-        len = HvKEYS(hash);
+        HE *he;
         hv_iterinit(hash);
-        for (i = 0; i < len; i++) {
-            HE *he = hv_iternext(hash);
+        while ((he = hv_iternext(hash))) {
             SV *val = HeVAL(he);
             if (val)
                 dump_prewalk(dumper, val);
@@ -608,7 +651,7 @@ dump_node(perl_yaml_dumper_t *dumper, SV *node)
 {
     yaml_char_t *anchor = NULL;
     yaml_char_t *tag = NULL;
-    char *class = NULL;
+    const char *class = NULL;
 
     if (SvTYPE(node) == SVt_PVGV) {
         SV **svr;
@@ -652,6 +695,15 @@ dump_node(perl_yaml_dumper_t *dumper, SV *node)
             }
             dump_scalar(dumper, node, tag);
         }
+#if PERL_REVISION > 5 || (PERL_REVISION == 5 && PERL_VERSION >= 11)
+        else if (ref_type == SVt_REGEXP) {
+            yaml_char_t *tag = (yaml_char_t *)form(TAG_PERL_PREFIX "regexp");
+            class = sv_reftype(rnode, TRUE);
+                if (!strEQ(class, "Regexp"))
+                     tag = (yaml_char_t *)form("%s:%s", tag, class);
+            dump_scalar(dumper, node, tag);
+        }
+#endif
         else {
             printf(
                 "YAML::XS dump unhandled ref. type == '%d'!\n",
@@ -692,8 +744,8 @@ yaml_char_t *
 get_yaml_tag(SV *node)
 {
     yaml_char_t *tag;
-    char *class;
-    char *kind = "";
+    const char *class;
+    const char *kind = "";
     if (! (
         sv_isobject(node) ||
         (SvRV(node) && ( SvTYPE(SvRV(node)) == SVt_PVCV))
@@ -725,8 +777,7 @@ dump_hash(
     int len;
     AV *av;
     HV *hash = (HV *)SvRV(node);
-    len = HvKEYS(hash);
-    hv_iterinit(hash);
+    HE *he;
 
     if (!anchor)
         anchor = get_yaml_anchor(dumper, (SV *)hash);
@@ -741,16 +792,18 @@ dump_hash(
     yaml_emitter_emit(&dumper->emitter, &event_mapping_start);
 
     av = newAV();
-    for (i = 0; i < len; i++) {
-        HE *he = hv_iternext(hash);
+    len = 0;
+    hv_iterinit(hash);
+    while ((he = hv_iternext(hash))) {
         SV *key = hv_iterkeysv(he);
         av_store(av, AvFILLp(av)+1, key); /* av_push(), really */
+        len++;
     }
     STORE_HASH_SORT;
     for (i = 0; i < len; i++) {
         SV *key = av_shift(av);
         HE *he  = hv_fetch_ent(hash, key, 0, 0);
-        SV *val = HeVAL(he);
+        SV *val = he ? HeVAL(he) : NULL;
         if (val == NULL) { val = &PL_sv_undef; }
         dump_node(dumper, key);
         dump_node(dumper, val);
@@ -809,6 +862,7 @@ dump_scalar(perl_yaml_dumper_t *dumper, SV *node, yaml_char_t *tag)
         plain_implicit = quoted_implicit = 1;
     }
 
+    SvGETMAGIC(node);
     if (!SvOK(node)) {
         string = "~";
         string_len = 1;
@@ -825,7 +879,7 @@ dump_scalar(perl_yaml_dumper_t *dumper, SV *node, yaml_char_t *tag)
         style = YAML_PLAIN_SCALAR_STYLE;
     }
     else {
-        string = SvPV(node, string_len);
+        string = SvPV_nomg(node, string_len);
         if (
             (string_len == 0) ||
             strEQ(string, "~") ||
@@ -833,16 +887,16 @@ dump_scalar(perl_yaml_dumper_t *dumper, SV *node, yaml_char_t *tag)
             strEQ(string, "false") ||
             strEQ(string, "null") ||
             (SvTYPE(node) >= SVt_PVGV) ||
-            ( !SvNIOK(node) && looks_like_number(node) )
+            ( dumper->quote_number_strings && !SvNIOK(node) && looks_like_number(node) )
         ) {
             style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
         }
         if (!SvUTF8(node)) {
-	    /* copy to new SV and promote to utf8 */
-	    SV *utf8sv = sv_mortalcopy(node);
+        /* copy to new SV and promote to utf8 */
+        SV *utf8sv = sv_mortalcopy(node);
 
-	    /* get string and length out of utf8 */
-	    string = SvPVutf8(utf8sv, string_len);
+        /* get string and length out of utf8 */
+        string = SvPVutf8(utf8sv, string_len);
         }
     }
     yaml_scalar_event_initialize(
@@ -949,7 +1003,7 @@ dump_ref(perl_yaml_dumper_t *dumper, SV *node)
 }
 
 int
-append_output(void *yaml, unsigned char *buffer, unsigned int size)
+append_output(void *yaml, unsigned char *buffer, size_t size)
 {
     sv_catpvn((SV *)yaml, (const char *)buffer, (STRLEN)size);
     return 1;
